////////////////////////////////////////////////////////////////////////
//                                                                    //
// convert.c                                                       //
//                                                                    //
// the implementation of the conversion class, it provides members    //
// for creating and converting a SIGMA input model.                   //
//                                                                    //
// (c)  1995 A.R. van der Valk, A. Eliens                              //
//                                                                    //
////////////////////////////////////////////////////////////////////////

#include <hush/dv/sigma/convert.h>
#include <hush/dv/sigma/sigma.h>

#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#define  MAX_PRIORITY 8


/***********************************************************************
                        HELP FUNCTIONS
***********************************************************************/
// returns the maximum number of edge conditions that occurs in
// a SIGMA input-model

int convert::mode = 0;


int
convert::max_nr_edge_cond(model *m)
{
    vertex_l_item *pt = m->first_vertex();
    edge_l_item *edge_pt = NULL;
    int teller = 0;
    int maximum = 0;
    while (pt != NULL)
    {
        edge_pt = m->first_edge();
        while (edge_pt != NULL)
        {
            // found an edge condition for this vertex
            if (strcmp(m->edge_origin(edge_pt), m->vertex_name(pt)) == 0)  
                teller++;
        
            edge_pt = m->next_edge(edge_pt);
        }
        if (teller > maximum)
            maximum = teller;

        teller = 0;
        pt = m->next_vertex(pt);
    }
    return (maximum);
}

/***********************************************************************
***********************************************************************/


// constructor for the convert class

convert::convert(char* filename) 
{
    _filename = filename;
    md = new model();
}


// the actual conversion from a SIGMA input-file to sim classes

convert::do_conversion(char *input_file, int _mode)
{
    convert::mode = _mode; // AE the various options and possibilities

    int nr_of_params = 0;
    int index = 0;
    char parameter[10];
    parameter[0] = 'i';
    parameter[1] = 'j';
    parameter[2] = 'k';
    parameter[3] = 'l';
    parameter[4] = 'm';
    parameter[5] = 'n';
    parameter[7] = 'o';
    parameter[8] = 'p';
    parameter[9] = 'q';


    // read the input file and open (if possible)  the output file

    md->read_model(input_file);

    //if (!_filename) _filename = md->def_outputfile();

    ostream* ofilep;
    
    if (_filename) {
	cerr << "output file is : " << _filename << endl;

    	ofstream ofilex(_filename, ios::out);
	ofilep = &ofilex;
    	if (!ofilep) // opened failed
    	{
        	cerr << "cannot open " << _filename << " for output\n"; 
        	exit(0);
    	}
    } else {
	ofilep = &cout;
    }

    ostream& ofile = *ofilep;
    // start with the includes and the declarations of the variables

    ofile << "/* This is the C++ program generated with the use of the input-file " <<  input_file << " */" << endl;

    ofile << endl << endl << "#include <hush/sigma.h>" << endl;
  
    if (md->def_ending_cond() == STOP_ON_EVENT)
    {
        ofile << endl << "/* STOP EVENT : " << md->def_stop_event() << " */" << endl;
        ofile << endl << "/* NR OF TIMES : " << md->def_end_cond_nr() << " */" << endl;
    }
    else if (md->def_ending_cond() == STOP_ON_TIME)
    {
        ofile << endl << "/* SIMULATION WILL QUIT AFTER : " << md->def_end_cond_nr() << " TIME UNITS  */" << endl;

    }
    
    ofile << endl << "simulation *sim;" << endl;
    ofile << "generator *gen;" << endl;

    // each traced variable gets a time average and a frequency average
    // histogram

    sigma_list_item *var = md->get_f_trace_var();
    while (var != NULL)
    {
        ofile << "histogram *hf" << md->stripped_trace_var(var) << ";" << endl;
        ofile << "histogram *hw" << md->stripped_trace_var(var) << ";" << endl;
        ofile << "analysis *af" << md->stripped_trace_var(var) << ";" << endl;
        ofile << "analysis *aw" << md->stripped_trace_var(var) << ";" << endl;
        var = md->get_n_trace_var(var);
    }

    // declaration of the edge condition array

    ofile << endl;
    ofile << "int edge_condition[" << max_nr_edge_cond(md) << "];" << endl;
    ofile << endl;

    // declaration of the global variables

    ofile << "/*    TOESTANDSVARIABELEN  */" << endl << endl;
    st_var_l_item *stvar = md->first_st_var();
    while (stvar != NULL)
    {
        switch(md->state_var_type(stvar))
        {
            case INTEGER:
                ofile << "int ";
                if ((index = md->is_array(stvar)) == 0)
                    ofile << md->state_var_name(stvar) << ";" <<endl;
                else
                    ofile << "*" << md->state_var_name(stvar) << " = new int[" << index << "];" << endl; 

                break;
            case REAL:
                ofile << "double ";
                if ((index = md->is_array(stvar)) == 0)
                    ofile << md->state_var_name(stvar) << ";" << endl;
                else
                    ofile << "*" << md->state_var_name(stvar) << " = new double[" << index << "];" << endl; 

                break;
            default:
                ofile << "Het type van " << md->state_var_name(stvar) << " is onbekend." << endl;
                break;
        }
        stvar = md->next_st_var(stvar);
    }

    // functions which determine if a variable is changed, so we know
    // when to sample

    ofile << endl << endl << endl;
    ofile << "int" << endl << "changed(int before, int after)" << endl;
    ofile << "{" << endl;
    ofile << "    if (before == after)" << endl << "        return (0);" << endl;
    ofile << endl << "    return (1);" << endl << "}";
    ofile << endl << endl;
    ofile << "int" << endl << "changed(double before, double after)" << endl;
    ofile << "{" << endl;
    ofile << "    if (before == after)" << endl << "        return (0);" << endl;
    ofile << endl << "    return (1);" << endl << "}";
    ofile << endl << endl;
    

    // declaration of the event types

    vertex_l_item *pt = md->first_vertex();
    st_var_l_item *p = NULL;
    sigma_list_item *st_var = NULL;

    while (pt != NULL)
    {
        ofile << "class " <<  md->vertex_name(pt) <<  " : public event" << " {";

        st_var = md->get_f_param(pt);
        nr_of_params = 0;

        if (st_var != NULL) 
        {
            ofile << endl << endl << "private: " << endl;
            while (st_var != NULL)
            {
                nr_of_params++;

                p = md->search_st_var(md->param_name(pt, st_var));
                if (p == NULL)
                    ofile << "Parameter unknown as state variable" << endl;
                else
                {
                    switch(md->state_var_type(p))
                    {
                        case INTEGER:
                            ofile << "    int ";
                            break;
                        case REAL:
                            ofile << "    double ";
                            break;
                        default:
                            ofile << "Unknown state variable type";
                            break;
                    }
                    ofile << md->state_var_name(p) << ";" << endl;
                    st_var = md->get_n_param(pt,st_var);
                }
            }
        } 


        ofile << endl << "public:" << endl << "    ";
        if (nr_of_params == 0)
            ofile << md->vertex_name(pt) << "();" << endl;
        else if (nr_of_params == 1)
        {
            ofile << md->vertex_name(pt) << "(";
            switch(md->state_var_type(p))
            {
                case INTEGER:
                    ofile << "int i);" << endl;;
                    break;
                case REAL:
                    ofile << "double i);" << endl;
                    break;
            }
        }
        else            
        {
            index = 0;
            ofile << md->vertex_name(pt) << "(";
            st_var = md->get_f_param(pt);
            while (st_var != NULL)
            {
                p = md->search_st_var(md->param_name(pt, st_var));
                switch(md->state_var_type(p))
                {
                    case INTEGER:
                        ofile << "int " << parameter[index];
                        break;
                    case REAL:
                        ofile << "double " << parameter[index];
                        break;
                }
                if (--nr_of_params == 0)
                    ofile << ");" << endl;
                else
                    ofile << ", ";

                index++;
                st_var = md->get_n_param(pt, st_var);
            }
        }

        ofile << "    virtual int operator()();" << endl; 
        ofile << "};" << endl << endl;
        pt = md->next_vertex(pt);
    }

    // implementation of the event types

    pt = md->first_vertex();
    ofile << endl << endl << endl;
    while (pt != NULL)
    {

        bool stop_event = false;
        index = 0;
        nr_of_params = 0;

        // determine stop condition

        if (md->def_ending_cond() == STOP_ON_EVENT)
            if (strcmp(md->def_stop_event(), md->vertex_name(pt)) == 0)
                stop_event = true;

        ofile << md->vertex_name(pt) << "::" << md->vertex_name(pt);
        st_var = md->get_f_param(pt);

        if (st_var == NULL)  // no parameters
        {
            ofile << "() : event(";
            if (stop_event == true)
                ofile << "0,0,1)" << endl << "{" << endl << "}" << endl << endl;
            else
                ofile << ")" << endl << "{" << endl << "}" << endl << endl;
        }
        else  // the constructor needs a parameter
        {
            ofile << "(";
            while (st_var != NULL)
            {
                p = md->search_st_var(md->param_name(pt, st_var));
                switch(md->state_var_type(p))
                {
                    case INTEGER:
                        ofile << "int " << parameter[index];
                        break;
                    case REAL:
                        ofile << "double " << parameter[index];
                        break;
                }
                index++;
                nr_of_params++;
                st_var = md->get_n_param(pt, st_var);
                if (st_var == NULL) // dus geen parameters meer
                {
                    ofile << ")" << " : event(";
                    if (stop_event == true)
                        ofile << "0,0,OK)" << endl << "{" << endl;
                    else
                        ofile << ")" << endl << "{" << endl;    
                }
                else
                    ofile << ", ";
            }
            
            index = 0;
            st_var = md->get_f_param(pt);
            while (nr_of_params > 0) 
            {
                ofile << "     ::" << md->param_name(pt, st_var) << " = " << md->param_name(pt,st_var) << " = ";
                ofile << parameter[index] << ";" << endl;
                index++;
                nr_of_params--;
                st_var = md->get_n_param(pt, st_var);
            }
            ofile << "}" << endl << endl;
        }

        // the operator

        ofile << endl << endl;
        ofile << "int " << md->vertex_name(pt) << "::operator()()" << endl << "{" << endl << endl;


	{ 
	char* p = md->vertex_action(pt);
	if (p) ofile << p << "; // action" << endl;
	}
        // handle the data sampling

        sigma_list_item *trace_var = md->get_f_trace_var();
        st_var_l_item *sv = NULL;
        char *hulp = NULL;

        if (pt != md->first_vertex())
        {
            while (trace_var != NULL)
            {
                if ((hulp = strstr(md->trace_var_name(trace_var),"[")) != NULL)
                {
                     // strippen van het "[" - je
                    int lengte = strlen(hulp);
                    char *stripped = new char[1024];
                    strncpy(stripped, md->trace_var_name(trace_var), strlen(md->trace_var_name(trace_var)) - lengte);

                    sv = md->search_st_var(stripped);
                }
                else
                    sv = md->search_st_var(md->trace_var_name(trace_var));

                if (sv == NULL)
                    cerr << "traced variable " << md->trace_var_name(trace_var) << " not declared as state variable." << endl;

                else
                {
                    ofile << "    hw" << md->stripped_trace_var(trace_var) << "->sample((double) " << md->trace_var_name(trace_var) << ",1.0);" << endl;
                    switch (md->state_var_type(sv))
                    {
                        case INTEGER:
                            ofile << "    int ";
                            break;
                        case REAL:
                            ofile << "    double ";
                            break;
                    }
                    ofile << "old" << md->stripped_trace_var(trace_var) << " = " << md->trace_var_name(trace_var) << ";" << endl;

                }
                trace_var = md->get_n_trace_var(trace_var);
            }
        }
        ofile << endl;


        // the state changes

        sigma_list_item *st_change = md->get_f_st_change(pt);
        while (st_change != NULL)
        {
            ofile << "    " << md->stripped_st_ch(pt, st_change) << ";" << endl;
            st_change = md->get_n_st_change(pt,st_change);
        }


        // if variables have changed, sample data

        ofile << endl;
        trace_var = md->get_f_trace_var();
        while (trace_var != NULL)
        {
                st_change = md->get_f_st_change(pt);
                while (st_change != NULL)
                {
                    if (strstr(md->st_change_name(pt, st_change), md->trace_var_name(trace_var)) != 0)
                    {

                        ofile << "    if (changed(old" << md->stripped_trace_var(trace_var) << ", " << md->trace_var_name(trace_var) << ") == 1)" << endl;
                        ofile << "        hf" << md->stripped_trace_var(trace_var) << "->sample((double) " << md->trace_var_name(trace_var) << ",1.0);" << endl;
                    }
                    st_change = md->get_n_st_change(pt,st_change);
                }
                trace_var = md->get_n_trace_var(trace_var);
         }


        // the edge conditions

        edge_l_item *edge_pt = md->first_edge();
        index = 0;
        ofile << endl;
        while (edge_pt != NULL)
        {
            if (strcmp(md->edge_origin(edge_pt), md->vertex_name(pt)) == 0)  
            {
                ofile << "    edge_condition[" << index << "] = " << "(" << md->edge_condition(edge_pt) << ");" << endl;
                index++;
            }
            edge_pt = md->next_edge(edge_pt);
        }
       
        edge_pt = md->first_edge();
        index = 0;
        char *buf = new char[1024];
        while (edge_pt != NULL)
        {
            if (strcmp(md->edge_origin(edge_pt), md->vertex_name(pt)) == 0) 
            {
                ofile << endl << "    if (edge_condition[" << index << "])" << endl;
                ofile << "    {" << endl;
                {
		char* p = md->edge_action(edge_pt);
		if (p) ofile << "	" << p << "; // action" << endl;
		} // create new one
                ofile << "        " << md->edge_destination(edge_pt) << " *";
                //strncpy(buf,md->edge_destination(edge_pt),2); buf[2] = '\0';
		sprintf(buf,"instance_of_%s",md->edge_destination(edge_pt));
                ofile << buf;
                ofile << ";" << endl << endl;
                ofile << "        " << buf << " = new " << md->edge_destination(edge_pt) << "(";

                sigma_list_item *attrib = md->get_f_attr(edge_pt);
                if (attrib == NULL)  // geen attributes
                    ofile << ");" << endl;
                else // there are attribute(s)
                {
                    while (attrib != NULL)
                    { 
                        ofile << md->attrib_name(edge_pt,attrib);
                        attrib = md->get_n_attr(edge_pt,attrib);
                        if (attrib == NULL) // no more attributes
                            ofile << ");" << endl;
                        else
                            ofile << ",";
                    }
                }

                // make priority opposite for use with sim

                ofile << "        " << buf << "->schedulingpriority(" << "(-1)*(" << md->edge_priority(edge_pt) << "));" << endl;
                ofile << "        sim->schedule(" << buf << ", " << md->edge_delay(edge_pt) << ");" << endl << "    }" << endl;
                
                ofile << endl;
                index++;
            }
            edge_pt = md->next_edge(edge_pt);
        }
 
        ofile << endl << "    sim->terminate(this);" << endl;
        ofile << "    return (OK);" << endl;
        ofile << "}" << endl << endl;
        
        pt = md->next_vertex(pt);
    }

/* declarations for the application constructor and main program */
 
    ofile << "class application : public session" << endl << "{" << endl;
    ofile << "public:" << endl;
    ofile << "    application(int argc, char** argv);" << endl;
    ofile << "    virtual int main();" << endl;
    ofile << "};" << endl << endl;
 
/* implementation of these functions */
 
    ofile << "application::application(int argc, char** argv) : session(argc, argv, \"" << md->def_outputfile() << "\")" << endl << "{ depends(0); " << endl << "}" << endl;
    
 
    ofile << endl << endl << "int application::main()" << endl <<"{" << endl; 
    ofile << endl << "    sim = new simulation();" << endl << endl;
    
    var = md->get_f_trace_var();
    while (var != NULL)
    {
        ofile << "    hf" << md->stripped_trace_var(var) << " = new histogram(\".hf" << md->stripped_trace_var(var) << "\",\"-type frequency -form statistics -title hf" << md->stripped_trace_var(var) << "\");" << endl;
 
        ofile << "    hw" << md->stripped_trace_var(var) << " = new histogram(\".hw" << md->stripped_trace_var(var) << "\",\"-type weighted -form statistics -title hw" << md->stripped_trace_var(var) << "\");" << endl;
 
        ofile << endl;

        ofile << "    tk -> pack(hf" << md->stripped_trace_var(var) << ");" << endl;
        ofile << "    tk -> pack(hw" << md->stripped_trace_var(var) << ");" << endl;
 
        ofile << endl;

        ofile << "    af" << md->stripped_trace_var(var) << " = new analysis(hf" << md->stripped_trace_var(var) << ");" << endl;
        ofile << "    aw" << md->stripped_trace_var(var) << " = new analysis(hw" << md->stripped_trace_var(var) << ");" << endl;
        var = md->get_n_trace_var(var);
    }
 
    // allocate the generator with the given seed

    ofile << endl << endl << "    gen = new generator(" << md->def_seed() << "," << md->def_seed() + 1 << "," << md->def_seed() + 2 << ");" << endl << endl;

     
    char *buf = NULL;
    buf = new char[3];
    pt = md->first_vertex();
    sigma_list_item *val_pt = md->get_f_init_val();

    //strncpy(buf, md->vertex_name(pt), 2); buf[2] = '\0';
    sprintf(buf,"instance_of_%s",md->vertex_name(pt));

    ofile << "    " << md->vertex_name(pt) << " *" << buf << " = new " << md->vertex_name(pt) << "(";
    
    if (val_pt == NULL)
        ofile << ");" << endl;
    else
    {
        while (val_pt != NULL)
        {
            ofile << md->get_value(val_pt);
            if ((val_pt = md->get_n_init_val(val_pt)) == NULL)
                ofile << ");" << endl;
            else
                ofile << ", ";
        }
    }
 
    // schedule first event at 0.0

    ofile << endl << "    sim->schedule(" << buf << ",0.0);" << endl;

    // handle ending condition

    switch(md->def_ending_cond())
    {
        case STOP_ON_EVENT:
            ofile << "    sim->run(" << md->def_end_cond_nr() << ");" << endl << endl;
            break;
        case STOP_ON_TIME:
            ofile << "    sim->run((double) " << md->def_end_cond_nr() << ");" << endl << endl;
            break;
    }


    // display the histogram on the screen

    var = md->get_f_trace_var();
    while (var != NULL)
    {
        ofile << "    cout << *hf" << md->stripped_trace_var(var) << ";" << endl;    
        ofile << "    cout << *hw" << md->stripped_trace_var(var) << ";" << endl;    
        ofile << "    af" << md->stripped_trace_var(var) << " -> confidence(\"percent 95\");" << endl;
        ofile << "    aw" << md->stripped_trace_var(var) << " -> confidence(\"percent 95\");" << endl;
        var = md->get_n_trace_var(var);
    }
    
    // delete the sim object

    ofile << endl << "    delete sim;" << endl;
    var = md->get_f_trace_var();
    while (var != NULL)
    {
        ofile << "    delete hf" << md->stripped_trace_var(var) << ";" << endl;
        ofile << "    delete hw" << md->stripped_trace_var(var) << ";" << endl;
        ofile << "    delete af" << md->stripped_trace_var(var) << ";" << endl;
        ofile << "    delete aw" << md->stripped_trace_var(var) << ";" << endl;
        var = md->get_n_trace_var(var);
    }

    ofile << "    tk->pack(\"quit\");" << endl;

    ofile << "}" << endl;

    /* the main program */
 
    ofile << endl << "main(int argc, char** argv)" << endl << "{" << endl;
    ofile << "    session *s;" << endl << endl;
    ofile << "    s = new application(argc,argv);" << endl;
    ofile << "    s -> run();" << endl;
    ofile << "    delete s;" << endl;
    ofile << "    session::statistics();" << endl;
    ofile << "    exit(0);" << endl << "}" << endl;
    //ofile.close();

    return(1); 
}



// destructor 

convert::~convert()
{
    md->~model();
}
