//
//  Vrml 97 library
//  Copyright (C) 1998 Chris Morley
//
//  %W% %G%
//  VrmlNode.h

#ifndef  _VRMLNODE_
#define  _VRMLNODE_

#include "config.h"
#include "System.h"		// error

// hash_map would be better, but it isn't portable yet...
#include _MAP_H_
#include _MULTIMAP_H_

#include <string.h>

NAMESPACE;

class Route;
class Viewer;

class VrmlNodeType;
class VrmlField;
class VrmlScene;


class VrmlNode {
  friend ostream& operator<<(ostream& os, const VrmlNode& f);

public:

  // Define the fields of all built in VrmlNodeTypes
  static VrmlNodeType *defineType(VrmlNodeType *t);
  virtual VrmlNodeType *nodeType() const;

  // VrmlNodes are reference counted, optionally named objects
  // The reference counting is manual (that is, each user of a
  // VrmlNode, such as the VrmlMFNode class, calls reference()
  // and dereference() explicitly). Should make it internal...

  VrmlNode ( VrmlScene *s = 0 );
  VrmlNode ( const VrmlNode & );
  virtual ~VrmlNode() = 0;

  // Copy the node.
  virtual VrmlNode *clone() const { return 0; }; //= 0; until all are implemented

  // Add/remove references to a VrmlNode. This is silly, as it
  // requires the users of VrmlNode to do the reference/derefs...
  VrmlNode *reference();
  void dereference();

  // Node type tests (should they move to the VrmlNodeType class?).
  // Remember to also add new ones to VrmlNodeProto.
  virtual bool isAnchorNode() const;
  virtual bool isAppearanceNode() const;
  virtual bool isAudioClipNode() const;
  virtual bool isChildNode() const;
  virtual bool isColorNode() const;
  virtual bool isCoordinateNode() const;
  virtual bool isFontStyleNode() const;
  virtual bool isGeometryNode() const;
  virtual bool isGroupNode() const;
  virtual bool isLightNode() const;
  virtual bool isScopedLightNode() const;
  virtual bool isProtoNode() const;
  virtual bool isMaterialNode() const;
  virtual bool isMovieTextureNode() const;
  virtual bool isNormalNode() const;
  virtual bool isScriptNode() const;
  virtual bool isSoundSourceNode() const;
  virtual bool isTextureCoordinateNode() const;
  virtual bool isTextureNode() const;
  virtual bool isTextureTransformNode() const;
  virtual bool isTouchSensorNode() const;

  // Node DEF/USE/ROUTE name
  void setName(const char *nodeName);
  static VrmlNode *lookup(const char *nodeName);

  const char *name() const;

  // Add to a scene. A node can belong to at most one scene for now.
  // If it doesn't belong to a scene, it can't be rendererd.
  virtual void addToScene( VrmlScene * );

  // Write self
  ostream& print(ostream& os, int indent) const;
  virtual ostream& printFields(ostream& os, int indent);
  static  ostream& printField(ostream&, int, const char*, const VrmlField&);

  // Indicate that the node state has changed, need to re-render
  void setModified();
  void clearModified() { d_modified = false; }
  virtual bool isModified() const;

  // Add a ROUTE from a field in this node
  void addRoute(const char *fromField, VrmlNode *toNode, const char *toField);

  // Delete a ROUTE from a field in this node
  void deleteRoute(const char *fromField, VrmlNode *toNode, const char *toField);

  // Pass a named event to this node. This method needs to be overridden
  // to support any node-specific eventIns behaviors, but exposedFields
  // (should be) handled here...
  virtual void eventIn(double timeStamp,
		       const char *eventName,
		       const VrmlField *fieldValue);

  // Set a field by name (used by the parser, not for external consumption).
  virtual void setField(const char *fieldName,
			const VrmlField &fieldValue);

  // Do nothing. Renderable nodes need to redefine this.
  virtual void render(Viewer *);

  // Do nothing. Grouping nodes need to redefine this.
  virtual void accumulateTransform(VrmlNode*);

  virtual VrmlNode* getParentTransform();

  virtual void renderAccumulatedTransforms(Viewer *);


protected:

  enum { INDENT_INCREMENT = 4 };

  // Send a named event from this node.
  void eventOut(double timeStamp,
		const char *eventName,
		const VrmlField &fieldValue);

  // Scene this node belongs to
  VrmlScene *d_scene;

  // True if a field changed since last render
  bool d_modified;

private:

  // Default constructor private to force use of (VrmlScene*) constructor.
  VrmlNode::VrmlNode();

  int d_refCount;		// Number of active references

  struct stringLess {
    bool operator()(const char* s1, const char* s2) const
    {
      return strcmp(s1, s2) < 0;
    }
  };

  typedef map <const char *, VrmlNode *, stringLess> NameMap;

  // Index of this node's name in the map
  NameMap::iterator d_nameIndex;

  // Map from DEF/USE names to nodes. Needs to be a stack of these
  // for managing PROTO namespaces (a DEF'd name inside a PROTO
  // is not available outside that PROTO implementation, and DEF'd
  // names outside a PROTO may not be USEd inside a PROTO)...
  // Also, this should be a per scene attribute, not a global one...
  // Static member variables are evil.
  static NameMap d_namedNodes;

  // Find routes given fieldnames
  typedef multimap <const char *, Route *, stringLess> RouteMap;

  // Routes from this node
  RouteMap d_routes;

};


// Routes
class Route {
public:
    char *fromEventOut;
    VrmlNode *toNode;
    char *toEventIn;
};


// Ugly macro used in printFields
#define PRINT_FIELD(_f) printField(os,indent+INDENT_INCREMENT,#_f,d_##_f)

// Ugly macros used in setField

#define ERROR theSystem->error

#define TRY_FIELD(_f,_t) \
(strcmp(fieldName, #_f) == 0) {\
    if ( fieldValue.to##_t() )\
      d_##_f = (Vrml##_t &)fieldValue;\
    else \
      ERROR("Invalid type (%s) for %s field of %s node (expected %s).\n",\
	    fieldValue.fieldTypeName(), #_f, nodeType()->getName(), #_t);\
  }

// For SFNode fields
#define TRY_SFNODE_FIELD(_f,_n) \
(strcmp(fieldName, #_f) == 0) { \
    VrmlSFNode *x=(VrmlSFNode*)&fieldValue; \
    if (fieldValue.toSFNode() && ((!x->get()) || x->get()->is##_n##Node())) \
      { d_##_f = (VrmlSFNode &)fieldValue; \
        if (d_scene && x->get()) x->get()->addToScene(d_scene); } \
    else \
      ERROR("Invalid type (%s) for %s field of %s node (expected %s).\n",\
	    fieldValue.fieldTypeName(), #_f, nodeType()->getName(), #_n);\
  }


#endif _VRMLNODE_

