//
//  Vrml 97 library
//  Copyright (C) 1998 Chris Morley
//  Copyrights on some portions of the code are held by others as documented
//  in the code. Permission to use this code for any purpose is granted as 
//  long as all other copyrights in the code are respected and this copyright
//  statement is retained in the code and accompanying documentation both 
//  online and printed.
//

#ifndef _VRMLNODETYPE_
#define _VRMLNODETYPE_
/**************************************************
 * VRML 2.0 Parser
 * Copyright (C) 1996 Silicon Graphics, Inc.
 *
 * Author(s)    : Gavin Bell
 *                Daniel Woods (first port)
 **************************************************
 */

//
// The VrmlNodeType class is responsible for storing information about node
// or prototype types.
//

#include "config.h"
#include _LIST_H_

NAMESPACE;

#include "VrmlField.h"
#include "VrmlMFNode.h"

class Doc;
class VrmlNode;
class VrmlScene;


class VrmlNodeType {
public:
  // Constructor.  Takes name of new type (e.g. "Transform" or "Box")
  // Copies the string given as name.
  VrmlNodeType(const char *nm,
	       VrmlNode* (*creator)(VrmlScene *scene) = 0);

  // Deallocate storage for name and PROTO implementations
  ~VrmlNodeType();

  VrmlNodeType *reference() { ++d_refCount; return this; }
  void dereference() { if (--d_refCount==0) delete this; }

  // Create a node of this type in the specified scene
  VrmlNode *newNode( VrmlScene* s = 0 ) const;

  // Routines for adding/getting eventIns/Outs/fields to this type
  void addEventIn(const char *name, VrmlField::VrmlFieldType type);
  void addEventOut(const char *name, VrmlField::VrmlFieldType type);
  void addField(const char *name, VrmlField::VrmlFieldType type,
		VrmlField *defaultVal = 0);
  void addExposedField(const char *name, VrmlField::VrmlFieldType type,
		       VrmlField *defaultVal = 0);

  void setFieldDefault(const char *name, VrmlField *value);

  VrmlField::VrmlFieldType hasEventIn(const char *name) const;
  VrmlField::VrmlFieldType hasEventOut(const char *name) const;
  VrmlField::VrmlFieldType hasField(const char *name) const;
  VrmlField::VrmlFieldType hasExposedField(const char *name) const;

  VrmlField *fieldDefault(const char *name) const;

  // Set the URL to retrieve an EXTERNPROTO implementation from.
  void setUrl(VrmlField *url, Doc *relative = 0);

  // Add a node to a PROTO implementation
  void addNode(VrmlNode *implNode);
  
  // Add an IS linkage to one of the PROTO interface fields
  void addIS(const char *isFieldName,
	     const VrmlNode *implNode,
	     const char *implFieldName);

  // Replace the values in the implementation nodes that are
  // equivalenced (via IS) to the specified interface field.
  void setIS(const char *isFieldName, VrmlField *value) const;

  const char *getName() const { return d_name; }

  VrmlMFNode *getImplementationNodes();

  VrmlNode *firstNode();

  typedef struct {
    VrmlNode *node;
    char *fieldName;
  } NodeFieldRec;

  typedef list<NodeFieldRec*> ISMap;

  typedef struct {
    char *name;
    VrmlField::VrmlFieldType type;
    VrmlField *defaultValue;
    ISMap thisIS;
  } NameTypeRec;

  typedef list<NameTypeRec*> FieldList;

private:

  // Grab the implementation of an EXTERNPROTO
  void fetchImplementation();

  void add(FieldList &, const char *, VrmlField::VrmlFieldType);
  VrmlField::VrmlFieldType has(const FieldList &, const char *) const;

  int d_refCount;

  char *d_name;

  VrmlMFString *d_url;		// Where the EXTERNPROTO lives.
  Doc *d_relative;

  VrmlNodeType *d_proto;
  VrmlMFNode *d_implementation;	// The PROTO implementation nodes

  VrmlNode* (*d_creator)( VrmlScene* );	// Pointer to function to create instances

  // Fields defined for this node type
  FieldList eventIns;
  FieldList eventOuts;
  FieldList fields;

};
#endif // _VRMLNODETYPE_
